#!/usr/bin/env node

import { dirname, basename, join } from 'path';
import fs from 'fs-extra';
import chalk from 'chalk';
import { isPlainObject } from 'lodash-es';
import glob from 'fast-glob';
import Enquirer from 'enquirer';
import shell from 'shelljs';
import { fileURLToPath } from 'url';

const cwd = process.cwd();

// The base of this repo -- we're copying files from there.
//
const bBase = join(dirname(fileURLToPath(import.meta.url)), '../');

console.log('***', cwd, bBase);

const appMap = new Map();
const gConfig = {
    dot: true,
    onlyFiles: true,
    absolute: true,
    ignore: ['**/node_modules']
};

// Offer choice of apps to install
//
const apps = (await glob(join(bBase, 'apps/*'), { ...gConfig, onlyDirectories: true })).reduce((acc, appDir) => {
    const name = basename(appDir);
    acc.push({
        name,
        message: fs.readFileSync(join(appDir, 'README.md'), { encoding: 'utf8' })
    });
    return acc;
}, []);

const appName = await (new Enquirer.Select({
    name: 'appName',
    message: 'Pick an App, any App...',
    choices: apps
})).run();

// Set /actors paths.
//
await add(`actors/**/*.js`, { appName });
await add(`openApi/**/*.yml`, { appName });
await add(`.editorconfig`);
await add(`.gitignore`);
await add(`.yarnrc`);
await add(`LICENSE.md`);

// Clear fields that will be updated by the implementation later (see below).
//
await add(`package.json`, { formatter: content => {

    delete content.bin; // New app won't be a cli out-of-the-box

    return {
        ...content,

        "name": "",
        "version": "1.0.0",
        "private": true,
        "description": "",
        "bfApplication": true,
        "author": {
            "name": "",
            "email": "",
            "url": ""
        },
        "repository": {
            "type": "git",
            "url": "",
            "directory": ""
        },
        "homepage": "",
        "bugs": "",
        "dependencies": {}
    }
}});

let errored = false;
for(const targetStub of appMap.keys()) {
    const targetPath = join(cwd, targetStub);
    if(await fs.pathExists(targetPath)) {
        console.log(chalk.bgRed.white.bold(`\n Error: FILE ALREADY EXISTS \n Can't generate source file [${targetStub}] \n into ${targetPath} \n The idea is to run '> bf create' within an EMPTY directory, once. `));
        errored = true;
        break;
    }
}

// No conflicts. Notify, and write files if approved.
//
if(!errored) {
    const dict = Array.from(appMap.entries());

    // re #initial: https://github.com/enquirer/enquirer/issues/279#issuecomment-623461898
    //
    const adding = await (new Enquirer.MultiSelect({
        message: `The following folders/files will be added into ${bBase}.\n  Hit ENTER to add them, or CTRL-C to abort.`,
        choices: dict.map(target => {
            return {
                name: target[0],
                value: target[1]
            }
        }),
        initial: Array.from(appMap.keys()),
        result(names) { return this.map(names); } // get back k/v of selections
    })).run();

    for(let [ targetStub, content ] of Object.entries(adding)) {
        const targetPath = join(cwd, targetStub);
        await fs.ensureFile(targetPath);
        if(isPlainObject(content)) {
            content = JSON.stringify(content, null, 2);
        }
        await fs.writeFile(targetPath, content, { encoding: 'utf8' });
        console.log(chalk.bgGreenBright.black.bold(` Wrote file ${targetStub} `));
    }

    // Get the new package.json properly filled out.
    // NOTE: #validate function signature also defines (<value>, state, item, index) args, if you need those.
    //
    const packUp = new Enquirer.Snippet({
        name: 'username',
        message: 'Fill out the fields in package.json',
        fields: [{
            name: 'repo_name',
            message: `Name of this application`,
            validate: strLenValidator(2)
        }, {
            name: 'repo_description',
            message: `A useful description of your application`,
            validate: strLenValidator(10)
        }, {
            name: 'author_name',
            message: `Name of this application's author`,
            validate: strLenValidator(2)
        }, {
            name: 'author_email',
            message: `Author's email address`,
            validate(value) {
                // Very basic email check, not intended to do rigid validation.
                //
                if(!/.+@.+\..+/.test(value)) {
                    return packUp.styles.danger(`Please enter a valid email`);
                }
                return true;
            }
        }, {
            name: 'author_url',
            message: `#url of author's portfolio, website, etc. Optional`,
            initial: '',
            validate: () => true // Optional
        }],
        template: `{
    "name": "\${repo_name}",
    "description": "\${repo_description}",
    "author": {
        "name": "\${author_name}",
        "email": "\${author_email}",
        "url": "\${author_url}"
    }
}
`
    });

    // Jut a little shortcut for above #validate-rs
    //
    function strLenValidator(minLength) {
        return (value='') => {
            if(value.length < minLength) {
                return packUp.styles.danger(`Must be at least ${minLength} characters long`);
            }
            return true;
        }
    }

    // Get new values, add to current package.json, and write updated version.
    //
    const pUp = (await packUp.run()).values;
    const packPath = join(cwd, 'package.json');
    const pack = await fs.readJson(packPath);

    pack.name = pUp.repo_name;
    pack.description = pUp.repo_description;
    pack.author.name = pUp.author_name;
    pack.author.email = pUp.author_email;
    pack.author.url = pUp.author_url;

    await fs.writeFile(packPath, JSON.stringify(pack, null, 2), { encoding: 'utf8' });

    // Ensure that some shell commands exist: yarn pm2 bf
    //
    npmInstall('yarn');
    npmInstall('pm2');
    npmInstall('@bandf/bf', 'bf');

    // Guide if redis not installed
    //
    if(!shell.which('redis-cli')) {
        console.log(chalk.bgCyanBright.black.bold(`
Redis is not installed. You will need to install Redis for this application to work.
On OSX the simplest way is to install 'brew' and then install Redis.
Visit https://brew.sh/, run the one-liner install, and then 'brew install redis'.
Follow the instructions on how to have Redis start on boot.`));
        process.exit(0);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////

async function add(pattern, { appName:name, formatter=f=>f }={}) {
    const sourceBase = name ? join(bBase, 'apps', name) : bBase;
    for(const sourceFile of await glob(join(sourceBase, pattern), gConfig)) {
        const relativePath = sourceFile.replace(sourceBase, '');
        let content = await fs.readFile(sourceFile, { encoding: 'utf8' });
        try {
            content = JSON.parse(content);
        } catch (err) {}
        appMap.set(relativePath, formatter(content));
    }
}

function npmInstall(bin, check) {
    if(shell.which(check ?? bin)) {
        console.log(chalk.bgCyanBright.black.bold(` Installing ${bin} `));
        const child = shell.exec(`npm i -g ${bin}`, {
            async: true,
            verbose: true
        });
        child.stdout.pipe(process.stdout);
    }
}

